/**CFile***********************************************************************

   FileName    [booleanOp.c]

   PackageName [booleanOp]

   Synopsis    [Main program for performing simple Boolean operation and
     equivalence verification on combinational circuits.]

   Description [Main program for performing simple Boolean operation  and
     equivalence verification on combinational circuits.
     It starts from a PLA description of the function on file.
     ]

   SeeAlso     []

   Author      [Stefano Quer]

   Copyright   [
    Copyright (c) May 19, 2003 by Politecnico di Torino.
    All Rights Reserved. This software is for educational purposes only.
    Permission is given to academic institutions to use, copy, and modify
    this software and its documentation provided that this introductory
    message is not removed, that this software and its documentation is
    used for the institutions' internal research and educational purposes,
    and that no monies are exchanged. No guarantee is expressed or implied
    by the distribution of this code.
    Send bug-reports and/or questions to: quer@polito.it
    ]

******************************************************************************/

#include "util.h"
#include "cuddInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

#define ROW_LEN_MAX 120
#define FAILURE 0
#define SUCCESS 1

/* Print Debugging Information while Running */
#define DEBUG_PRINTING 0

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int file2Bdd (DdManager *ddMgr, char *fileName, DdNode **function);
static int printStat (DdManager *ddMgr, DdNode *f);

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Main program for Boolean Operation.]

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

int
main(
  int  argc,
  char ** argv
  )
{
  int i;
  char name[ROW_LEN_MAX];
  DdManager *ddMgr;
  DdNode **function;

  /*
   *  Print Out Command and Version and etc.
   */

  if (argc != 2 && argc!=4) {
    fprintf (stdout, "Run Progam as:\n  ./booleanOp f\n  where f is a name of the file\n");
    fprintf (stdout, "or as:\n  ./booleanOp f1 op f2\n  where f1 a nd f2 are name of the file\n");
    fprintf (stdout, "  and op: e=equivalence verification, a=and, o=or, x=xor\n");
    fflush (stdout);
    exit (1);
  }

  fprintf (stdout, "# Running Date: ");
  fflush (stdout);
  system ("date");
  if (gethostname (name, ROW_LEN_MAX) == 0) {
    fprintf (stdout, "# HostName: %s\n", name);
  }
  fflush (stdout);

  fprintf (stdout, "# Command: ");
  for (i=0; i<argc; i++) {
    fprintf (stdout, "%s ", argv[i]);
  }
  fprintf (stdout, "\n");
  fprintf (stdout, "# CUDD Version: ");
  Cudd_PrintVersion (stdout);

  /*
   *  Init CUDD Manager
   */

  ddMgr = Cudd_Init (0, 0, CUDD_UNIQUE_SLOTS, 32768, 0);
  if (ddMgr == NULL) {
    return (FAILURE);
  }

  Cudd_EnableReorderingReporting (ddMgr);

  /*
   *  Read File and Create Function
   */

  if (argc == 2) {

    function = ALLOC (DdNode *, 1);

    file2Bdd (ddMgr, argv[1], &function[0]);

    printStat (ddMgr, function[0]);

    Cudd_RecursiveDeref (ddMgr, function[0]);

  } else {

    function = ALLOC (DdNode *, 3);

    file2Bdd (ddMgr, argv[1], &function[0]);

    fprintf (stdout, "---> First Function:\n");
    printStat (ddMgr, function[0]);

    file2Bdd (ddMgr, argv[3], &function[1]);

    fprintf (stdout, "---> Second Function:\n");
    printStat (ddMgr, function[1]);

    /*
     *  Perform Operation IFF Necessary
     */

    if (strcmp (argv[2], "e") == 0) {
      /* f1 == f2 ? */
      if (function[0] == function[1]) {
        fprintf (stdout, "---> Funtion are equivalent!\n");
      } else {
        fprintf (stdout, "---> Funtion are NOT equivalent!\n");
      }

      Cudd_RecursiveDeref (ddMgr, function[0]);
      Cudd_RecursiveDeref (ddMgr, function[1]);

    } else {

      if (strcmp (argv[2], "a") == 0) {
        /* f = f1 AND f2 */
        fprintf (stdout, "---> Perform AND\n");
        function[2] = Cudd_bddAnd (ddMgr, function[0], function[1]);
      }
      if (strcmp (argv[2], "o") == 0) {
        /* f = f1 OR f2 */
        fprintf (stdout, "---> Perform OR\n");
        function[2] = Cudd_bddOr (ddMgr, function[0], function[1]);
      }
      if (strcmp (argv[2], "x") == 0) {
        /* f = f1 XOR f2 */
        fprintf (stdout, "---> Perform XOR\n");
        function[2] = Cudd_bddXor (ddMgr, function[0], function[1]);
      }
    
      fprintf (stdout, "---> Result Function:\n");
      printStat (ddMgr, function[2]);

      Cudd_RecursiveDeref (ddMgr, function[0]);
      Cudd_RecursiveDeref (ddMgr, function[1]);
      Cudd_RecursiveDeref (ddMgr, function[2]);
    
    }   

  }

#if DEBUG_PRINTING
  Cudd_PrintDebug (ddMgr, finalRel, 0, 1);
#endif

#if DEBUG_PRINTING
  Cudd_PrintInfo (ddMgr, stdout);
#endif

  Cudd_Quit (ddMgr);

  return (SUCCESS);
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Read the File and Create the BDD.]

  Description [Read the File (containing a PLA description of a function)
   and Create the corresponding BDD.]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/

static int
file2Bdd (
  DdManager *ddMgr,
  char *fileName,
  DdNode **function
  )
{
  DdNode *var, *cube, *f, *tmpBdd;
  FILE *fp;
  int i, n, varN;
  char row[ROW_LEN_MAX];
 
  fp = fopen (fileName, "r");
  if (fp == NULL) {
    fprintf (stderr, "No File!\n");
    exit (1);
  }

  /*
   *  Read Number of Variable
   */
 
  if (fgets (row, ROW_LEN_MAX, fp) == NULL) {
    fprintf (stderr, "Wrong File Format!\n");
    exit (1);
  }
  n = sscanf (row, "%d", &varN);
  if (n==EOF) {
    fprintf (stderr, "Wrong File Format!\n");
    exit (1);
  }

  /*
   *  Read Cubes
   */

  f = Cudd_Not (DD_ONE (ddMgr));
  while (fgets (row, ROW_LEN_MAX, fp) != NULL) {
    cube = DD_ONE (ddMgr);

    for (i=0; i<strlen(row); i++) {
      if (row[i] == ' ') {
        break;
      }
      if (row[i] == '-') {
        continue;
      }
      var = Cudd_bddIthVar (ddMgr, i);  
      Cudd_Ref (var);
      if (row[i] == '1') {
        tmpBdd = Cudd_bddAnd (ddMgr, cube, var);   
      } else {
        tmpBdd = Cudd_bddAnd (ddMgr, cube, Cudd_Not (var));   
      }
      Cudd_Ref (tmpBdd);
      Cudd_RecursiveDeref (ddMgr, cube);
      cube = tmpBdd;
    }

    tmpBdd = Cudd_bddOr (ddMgr, f, cube);   
    Cudd_Ref (tmpBdd);
    Cudd_RecursiveDeref (ddMgr, f);
    f = tmpBdd;
  }

#if DEBUG_PRINTING
  fprintf (stdout, "Number of BDD var = %d\n", Cudd_ReadSize (ddMgr));
#endif

  *function = f;

  return (SUCCESS);
}

static int
printStat (
  DdManager *ddMgr,
  DdNode *f
  )
{
  if (f==NULL) {
    return (FAILURE);
  } else {
    fprintf (stdout, "BDD Size: %d\n", Cudd_DagSize (f));
    fprintf (stdout, "Number of Minterm: %d\n",
     ((int) Cudd_CountMinterm (ddMgr, f, Cudd_ReadSize (ddMgr))));
    fprintf (stdout, "Function: \n");
    Cudd_PrintMinterm (ddMgr, f);

    return (SUCCESS);
  }
}

